// This program is free software: you can use, modify and/or redistribute it
// under the terms of the simplified BSD License. You should have received a
// copy of this license along this program. If not, see
// <http://www.opensource.org/licenses/bsd-license.html>.
//
// Copyright (C) 2012, Jérôme Gilles <jegilles@math.ucla.edu>
// All rights reserved.

/**
 * @file MaoGilles.c
 * @brief Mao-Gilles stabilization algorithm demo for IPOL
 * @author Jérôme Gilles <jegilles@math.ucla.edu>
 *
 *
 * Copyright (c) 2012, Jérôme Gilles
 * All rights reserved.
 *
 * This program is free software: you can use, modify and/or
 * redistribute it under the terms of the simplified BSD License. You
 * should have received a copy of this license along this program. If
 * not, see <http://www.opensource.org/licenses/bsd-license.html>.
 */

/**
 * @mainpage
 * @verbinclude readme.txt
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#ifdef _OPENMP
#include <omp.h>
#endif

#include "iio.h"
#include "MaoGilles.c"



/**
 * @brief Mao-Gilles stabilization algorithm
 * @param arg1 the generic name of the input sequence
 * @param arg2 first frame index in the input sequence
 * @param arg2 last frame index in the input sequence
 * @param lambda the ROF regularization parameter
 * @param delta the Bregman gradient step
 *
 * This code provides an executable for the Mao-Gilles stabilization algorithm.
 *
 * The input sequence generic namehas the following form:
 * Assuming that the files of each input frame is named like "frame_000.png,
 * frame_001.png, ... then its corresponding generic name is "frame_%03d.png"
 * (it uses the C convention).
 * This executable uses the whole sequence to generate one restaured image
 * which is saved as "u.png"
 */


int
main(int argc, char **argv)
{

    float **seq=NULL;
    char name[1024];
    float *u=NULL;
		float *umean=NULL;
		float *vecx=NULL;
    float *vecy=NULL;
    float *f=NULL;
    int Ni,Nf,Nframes,n,m,i,width,height,Nbregman,Nsplitting,Saveflow;
    float lambda;
    float dt;
		
		clock_t start;
		clock_t end;

    //input parameters
    //1: filename pattern (ex: image_%04d.png)
    //2: index of first frame
    //3: index of last frame
    //4: lambda (regularization parameter)
    //5: dt (time step)
		//6: Nbregman (number of iteration for the Bregman loop)
		//7: Nsplitting (number of iteration for the Splitting loop)
		//8: Saveflow (activate or not the recording of flows)

    if (argc<9) {
        fprintf(stderr,"usage: MaoGilles image_generic_name first_index \
                last_index lambda delta Nbregman Nsplitting \
								saveflow_0_or_1\n\n");
        exit(EXIT_FAILURE);
    }

    // Get the input parameters
    Ni = atoi(argv[2]); //get the index of the first frame to load
    Nf = atoi(argv[3]); //get the index of the last frame to load
    Nframes = Nf-Ni+1; //compute the number of frames in the sequence
    lambda = atof(argv[4]); //regularization parameter
    dt = atof(argv[5]); //time step
		Nbregman = atoi(argv[6]); //number of iteration for the Bregman loop
		Nsplitting = atoi(argv[7]); //number of iteration for the Splitting loop
		Saveflow = atoi(argv[8]); //1 of you want to save the flows between the 
															//first restored frame and the first ten input 
															//frames, 0 otherwise

    fprintf(stderr,"Ni=%d - Nf=%d - Lambda=%f - dt=%f - Nbregman=%d - \
						Nsplitting=%d - Saveflow=%d\n",Ni,Nf,lambda,dt,Nbregman,\
						Nsplitting,Saveflow);

    //Memory Allocations
    seq	= (float **)xmalloc(sizeof(float *)*Nframes); //Original sequence

    //load sequence
    for (n=Ni; n<Nf+1; n++)
    {
        sprintf(name,argv[1],n);
        seq[n-Ni] = iio_read_image_float(name,&width,&height);
    }
    fprintf(stderr,"Sequence loaded\n");

    // Allocation of the output image
    u = (float *)xmalloc(sizeof(float)*width*height);
		umean = (float *)xmalloc(sizeof(float)*width*height);

    //Initialize u with the temporal mean
    temporal_mean(seq,u,Nframes,width,height);
		iio_save_image_float("Mean.png", u, width, height);

    // Perform the restoration process
		start=clock();
    MaoGilles_Stabilization(seq,u,width,height,Nframes,lambda,dt,Nbregman,\
														Nsplitting);
		end=clock();

    // Save the restored image
    iio_save_image_float("Restored.png", u, width, height);
    fprintf(stderr,"Image saved\n");
		fprintf(stderr,"Computed in %ld s\n",(end-start)/ CLOCKS_PER_SEC);

		// Save the optical flow between the first restored image and the
		// ten first input frames
		if (Saveflow==1) {
			uint8_t (**view)[3] = matrix_build(width, height, sizeof**view);
			vecx = (float *)xmalloc(sizeof(float)*width*height);
			vecy = (float *)xmalloc(sizeof(float)*width*height);
			f = xmalloc(2 * width * height * sizeof(float));

			for (m=0; m<(Nframes>10?10:Nframes); m++)
 			{
	   		optflow(u,seq[m],vecx,vecy,width,height);
     		for (i = 0; i < width * height; i++)
     		{
	       	f[2*i] = vecx[i];
         	f[2*i+1] = vecy[i];
     		}

				sprintf(name,"flow%03d.png",m+1);
				viewflow_flat(view[0][0], f, width, height, 1);
				iio_save_image_uint8_vec(name, view[0][0], width, height, 3);
   		}
	  	free(vecx);
			free(vecy);
			free(f);
		}
    // Memory freeing
    free(u);
    free(seq);

    exit(EXIT_SUCCESS);
}
